<?// -*-c-*- 
  // This is the PHP class Config::PHP
  // by Thomas Linden <tom@daemon.de>

  class Config {
     function Config($argument) {
        // new()
       if(is_string($argument)) {
	   // treat the var as a filename
	   // try to open it
	   $content = $this->open($argument);
	   // var_dump($content);
	   $this->config  = $this->parse(array(), $content);
	}
	elseif(is_array($argument)) {
	   // treat as an array of config lines
	   $content = $argument;
	   $config  = $this->parse(NULL, $content);
	}
	else {
	   // hmm, I mean this could be rubbish...
	   die("No supported argument to new()!\n");
	}
     }

     function getall() {
       // just return the resulting hash
       return $this->config;
     }

     function open($file) {
        // iterate over $content and look for C-comments, heredocs or 
	// include statements
       $content = array();
	if( file_exists($file) && is_readable($file) ) {
	   $lines = file($file) or die("Could not open configuration file $file!\n");
	}
	else {
	  die("Configuration file does not exist: $argument!\n");
	}
	while(list ($num, $line) = each($lines)) {
	   $line = trim($line);  // remove leading/trailing whitespaces
	   $line = preg_replace("/\n$/", "", $line);

	   if(!$hierend) {
	      if(preg_match("/^\s*#/", $line) || preg_match("/^\s*$/", $line)) {
	         // ignore comments and empty lines
	         continue;
	      }
	      if(preg_match("/#/", $line)) {
		if(preg_match("/\\#/", $line)) {
		  $line = preg_replace("/\\\\#/", chr(201), $line); // keep escaped #'s
		  $line = preg_replace("/#.+$/", "", $line);        // but remove everything after a non-escaped #
		  $line = ereg_replace(chr(201), "#", $line);
		}
		else {
		  $line = preg_replace("/#.+$/", "", $line);
		}
		$line = trim($line);
	      }
	   }
	   
	   if(preg_match("/^\s*(.+?)(\s*=\s*|\s+)<<(.+?)$/", $line, $match)) {
	      // we are @ the beginning of a here-doc
	      $hier    = $match[1]; // $hier is the actual here-doc
              $hierend = $match[3]; // the here-doc end string, i.e. "EOF"
	      $hierdoc = array();
	   }
	   elseif($hierend && preg_match("/^(\s*)$hierend$/", $line, $match)) {
	      // the current here-doc ends here
	      $hier .= "=" . join("\n", $hierdoc);  
	      array_push($content, $hier);  // push the complete here doc onto the content stack
	      $hier = NULL; $hierdoc = NULL; $hierend = NULL;  // reset heredoc variables
	   }
	   elseif(preg_match("/^\s*\/\*/", $line)) {
	      // the beginning of a C-comment ("/*"), from now on ignore everything.
	      if(preg_match("/\*\/\s*$/", $line)) {
	         // C-comment end is already there, so just ignore this line!
		 $c_comment = 0;
	      }
	      else {
	         $c_comment = 1;
	      }
	   }
	   elseif(preg_match("/\*\//", $line)) {
	      // C-Comment END
	      if (!$c_comment) {
	         die("invalid syntax: found end of C-comment without previous start(:$num)!\n");
	      }
	      $c_comment = 0; // the current C-comment ends here, go on 
	   }
	   elseif(preg_match("/\\\\$/", $line)) {
	      // a multiline option, indicated by a trailing backslash
	      $line = preg_replace("/\\\\$/", "", $line); // remove the trailing backslash
	      if(!$c_comment) {
	         $longline .= $line;
	      }
	   }
	   elseif(!$c_comment) {
	      // any "normal" config lines not inside C-comments
	      if($longline) {
	         // previous stuff was a longline and this is the last line of the longline
		 array_push($content, $longline . $line); // push it onto the content stack
		 $longline = NULL;
	      }
	      elseif($hier) {
	         // we are inside a here-doc
		 array_push($hierdoc, $line); // push onto here-doc stack
	      }
	      else {
	         if(preg_match("/^<<include (.+?)>>$/", $line, $match)) {
		    // include external config file
		    $this->open($match[1]); // call $this->open with the argument to include assuming it is a filename
		 }
		 else {
		    // standard config line, push it onto the content stack
		    array_push($content, $line);
		 }
	      }
	   }
	}
	return $content;
     }

     function parse($config, $content) {
       // print("config:\n" . var_dump($config) . "\ncontent:\n" . var_dump($content) . "\n"); 
        // actually parse the preprocessed contents
	while(list ($num, $line) = each($content)) {
	   list($option, $value) = preg_split("/\s*=\s*|\s+/", $line, 2);
	   // print("option: $option, value: $value\n");
	   if(preg_match("/^\"/", $value) || preg_match("/\"$/", $value)) {
	     // remove leading and trailing apostrophs
	     $line = preg_replace("/^\"/", "", $line);
	     $line = preg_replace("/\"$/", "", $line);
	     // print("line: $line\n");
	   }
	   if(!$block) {
	     // not inside a block @ the moment
	     if(preg_match("/^<([^\/]+?.*?)>$/", $line, $match)) {
	       // it is a block
	       $level++;
	       $block = $match[1];
               // is it a named block? IF yes, store the name separately
	       list($grab, $blockname) = preg_split("/\s\s*/", $block, 2);
	       if ($blockname) {
		 $block = $grab;
	       }
	       $newcontent = array();
	       continue; // the while loop
	     }
	     elseif(preg_match("/^<\/(.+?)>$/", $line, $match)) {
	       // it is an end block, but we dont have a matching block!
               die("EndBlock \"</$match[1]>\" has no StartBlock statement (level: $level, chunk $num)!\n");
	     }
	     else {
	       if(isset($config[$option])) {
		 // value exists more than once, make it an array
		 if(is_array($config[$option])) {
		   // it is already an array, so just push
		   array_push($config[$option], $value);
		 }
		 else {
		   // convert scalar to array
		   $savevalue = $config[$option];
		   $config[$option] = array();
		   array_push($config[$option], $savevalue, $value);
		 }
	       }
	       else {
		 // print("$option = $value\n");
		 $config[$option] = $value;
	       }
	     }
	   }
	   elseif(preg_match("/^<([^\/]+?.*?)>$/", $line)) {
             // found a start block inside a block, dont forget it
	     $block_level++;                 // $block_level indicates wether we are still inside a node
             array_push($newcontent, $line); // push onto new content stack For later recursive call of parse
	   }
	   elseif(preg_match("/^<\/(.+?)>$/", $line)) {
	     // endblock
	     if($block_level) {
	       // this endblock is not the one we are searching for, decrement and push
	       $block_level--;
	       array_push($newcontent, $line);
	     }
	     else {
	       if($blockname) {
		 // NAMED BLOCK
		 if(isset($config[$block][$blockname])) {
		   // the named block already exists, make it an array
		   $savevalue = $config[$block][$blockname];
		   $config[$block][$blockname] = NULL;
		   $ar = array(); // new empty array
		   // preserve array if any
		   array_push($ar, $savevalue);
		   array_push($ar, $this->parse(array(), $newcontent)); // append it
		   $config[$block][$blockname] = $ar;
		   $ar = NULL;
 		 }
		 else {
		   $config[$block][$blockname] = $this->parse($config[$block][$blockname], $newcontent);
		 }
	       }
	       else {
		 // STANDARD BLOCK
		  if(isset($config[$block])) {
		   // the block already exists, make it an array
		   $savevalue = $config[$block];
		   $config[$block] = array();
		   $ar = array(); // new empty array
		   // preserve array if any
		   array_push($ar, $savevalue);
		   array_push($ar, $this->parse(array(), $newcontent)); // append it
		   $config[$block] = $ar;
		   $ar = NULL;
 		 }
		 else {
		   $config[$block] = $this->parse($config[$block], $newcontent);
		 }
	       }
	     }
	     $blockname = NULL;
	     $block     = NULL;
	     $level--;
	     continue;
	   }
	   else {
	     // inside $block, just push onto new content stack
	     array_push($newcontent, $line);
	   }
	}
	if($block) {
          // $block is still defined, which means, that it had
          // no matching endblock!
	  die("Block \"<$block>\" has no EndBlock statement (level: $level, chunk $num)!\n");
	}
	return $config;
     }
  }
?>
